/* file.c - output an image as a TIFF or pnm file.
   $Id: file.c,v 1.4 1994/12/23 16:28:59 ceder Exp $
   Copyright (C) 1994  Lysator Computer Association

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdio.h>

#include <tiffio.h>

#include "scanner.h"

int
output_pnm(const char *filename,
	   int composition,
	   unsigned char *image,
	   long img_len,
	   long xpixels,
	   long ypixels)
{
  FILE *fp;

  if (verbose)
    fprintf(stderr, "Saving %ld bytes as %s to %s\n",
	    img_len, composition == 2 ? "PGM" : "PBM", 
	    filename == NULL ? "standard output" : filename);

  if (filename == NULL)
    {
      fp = stdout;
      filename = "standard output";
    }
  else
    if ((fp = fopen(filename, "wb")) == NULL)
      {
	fprintf(stderr, "opening %s", filename);
	perror(" for writing");
	return -1;
      }

  if (composition == 2)
    {
      unsigned char *p;
      unsigned char cvt[64];
      int ix;
      
      for (ix = 0; ix < 63; ix++)
	cvt[ix] = ((63-ix) << 2) + (((63-ix) & 0x30) >> 4);

      fprintf(fp, "P5 %ld %ld 255\n", xpixels, ypixels);

      /* Invert all pixels and convert to 8-bit PGM. */
      for (p = image; p < image + img_len; p++)
	*p = cvt[*p];
    }
  else
    fprintf(fp, "P4 %ld %ld\n", xpixels, ypixels);

  if (fwrite(image, 1, img_len, fp) != img_len)
    {
      fprintf(stderr, "error writing to ");
      perror(filename);
      return -1;
    }
    
  if (fclose(fp) != 0)
    {
      fprintf(stderr, "error closing ");
      perror(filename);
      return -1;
    }

  return 0;
}

#if 0
static void
packbits(unsigned char *base, long xpixels)
{
  unsigned char *dest = base;
  char tmp = 0;
  int bits = 0;
  
  while (xpixels-- > 0)
    {
      switch(bits)
	{
	case 0:
	  tmp = *base++ << 2;
	  break;
	case 2:
	  tmp |= *base++ & 0x3f;
	  *dest++ = tmp;
	  tmp = 0;
	  break;
	case 4:
	  tmp |= (*base >> 2) & 0x0f;
	  *dest++ = tmp;
	  tmp = (*base++ << 6) & 0xc0;
	  break;
	case 6:
	  tmp |= (*base >> 4) & 0x03;
	  *dest++ = tmp;
	  tmp = (*base++ << 4) & 0xf0;
	  break;
	default:
	  fprintf(stderr, "nasty internal error");
	  exit(1);
	}
      bits += 6;
      bits &= 7;
    }
  if (bits)
    *dest++ = tmp;
  if (dest > base)
    {
      fprintf(stderr, "poor fast scanf overpowered by mutant turtle power\n");
      exit(1);
    }
}
#endif

int
output_tiff(const char *filename,
	    int composition,
	    int separation,
	    unsigned char *image,
	    long img_len,	/* bytes */
	    long xpixels,
	    long ypixels,
	    long xresolution,	/* dpi */
	    long yresolution,	/* dpi */
	    float xorigin,	/* cm */
	    float yorigin)	/* cm */
{
  TIFF *tif;
  long block;
  long row;

  if (verbose)
    fprintf(stderr, "Saving %ld bytes as TIFF file %s\n",
	    img_len, filename);

  tif = TIFFOpen(filename, "w");
  if (tif == NULL)
    {
      fprintf(stderr, "TIFFOpen failed\n");
      return -1;
    }

  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, xpixels);
  TIFFSetField(tif, TIFFTAG_ORIENTATION, 1);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
  TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
  TIFFSetField(tif, TIFFTAG_XRESOLUTION, (float)xresolution);
  TIFFSetField(tif, TIFFTAG_YRESOLUTION, (float)yresolution);
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
  TIFFSetField(tif, TIFFTAG_SOFTWARE, version);
  TIFFSetField(tif, TIFFTAG_XPOSITION, xorigin * 2.54);
  TIFFSetField(tif, TIFFTAG_YPOSITION, yorigin * 2.54);
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  if (composition == 2)
    {
      unsigned char cvt[64];
      int ix;
      
      for (ix = 0; ix < 63; ix++)
	cvt[ix] = (ix << 2) + ((ix & 0x30) >> 4);

      TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
      TIFFSetField(tif, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
      TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
      for (row = 0; row < ypixels; row++)
	{
	  for (ix = 0; ix < xpixels; ix++)
	    image[row*xpixels+ix] = cvt[image[row*xpixels+ix]];

	  if (TIFFWriteScanline(tif, image + row * xpixels, row, 0) != 1)
	    {
	      fprintf(stderr, "TIFFWriteScanline failed on row %ld\n", row);
	      return -1;
	    }
	}
    }
  else
    {
      TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 1);
      TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
      block = xpixels / 8;
      if (separation == 0)
	if (composition == 0)
	  TIFFSetField(tif, TIFFTAG_THRESHHOLDING, THRESHHOLD_BILEVEL);
	else
	  TIFFSetField(tif, TIFFTAG_THRESHHOLDING, THRESHHOLD_BILEVEL);
      for (row = 0; row < ypixels; row++)
	{
	  if (TIFFWriteScanline(tif, image + row * block, row, 0) != 1)
	    {
	      fprintf(stderr, "TIFFWriteScanline failed on row %ld\n", row);
	      return -1;
	    }
	}
    }

  TIFFClose(tif);
  return 0;
}
