/* scanner - scan images from an M3191F2
   $Id: main.c,v 1.8 1994/07/14 23:02:24 ceder Exp $
   Copyright (C) 1994  Lysator Computer Association

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* First version written by ceder1994-06-22. */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

#include "scanner.h"
#include "getopt.h"

struct option longopts[] =
  {
    { "version",            no_argument, 0, 'V' },
    { "help",               no_argument, 0, 'h' },

    { "origo",              required_argument, 0, 'o' },
    { "rectangle",          required_argument, 0, 'r' },

    { "brightness",         required_argument, 0, 'b' },
    { "threshold",          required_argument, 0, 't' },
    { "image-composition",  required_argument, 0, 'c' },
    { "halftone",           required_argument, 0, 'p' },
    { "x-resolution",       required_argument, 0, 'X' },
    { "y-resolution",       required_argument, 0, 'Y' },

    { "outline-extraction", no_argument, 0, 'x' },
    { "edge-enhancement",   no_argument, 0, 'e' },
    { "separation",         no_argument, 0, 's' },
    { "mirror",             no_argument, 0, 'm' },

    { "verbose",            no_argument, 0, 'v' },

    { "tiff",               required_argument, 0, 'i' },

    { 0, 0, 0, 0 }
  };

int verbose = 0;

static char
usage[] = "Usage: %s [options] [filename]\nUse ``--help'' for help.\n";

static char *progname;

int
main(int argc,
     char **argv)
{
  int optc;
  /* User option: origin of image [cm]. */
  float xcm = 0;
  float ycm = 0;
  /* User option: size of image [cm] */
  float width = 21.6;
  float height = 29.6;
  /* User options: Functions that the scanner can perform (1=on, 0=off). */
  int outline_extraction = 0;
  int edge_enhancement = 0;
  int separation = 0;
  int mirroring = 0;
  /* User options: Parameters to the scanner (see scanner doc) */
  int brightness = 0;
  int threshold = 4;
  int composition = 0;
  int halftone = 0;
  int xresolution = 300;
  int yresolution = 300;
  /* Lookup table for brightness and threshold. */
  int lookup[8] = { 0, 0xff, 0xd8, 0xb4, 0x90, 0x6c, 0x48, 0x24 };
  /* User option: want a TIFF file */
  int use_tiff = 0;
  /* Number of errors found in user supplied options. */
  int err = 0;
  /* Communication with the scanner is done through fd. */
  int fd;
  /* The scanned image is stored here */
  char *image = 0;
  long img_len;
  long img_bufsize = 0;
  /* Actual size of scanned image in pixels. */
  long xpixels;
  long ypixels;

  progname = argv[0];
  
  while ((optc = getopt_long(argc, argv, "vho:r:b:t:c:p:xesmiX:Y:",
			     longopts, 0)) != EOF)
    {
      switch (optc)
	{
	case 'V':
	  puts(version);
	  exit(0);
	case 'v':
	  verbose++;
	  break;
	case 'h':
	  puts("scanf -- scan images from a Fujitsu M3191F2 "
	       "size A4 flatbed scanner.");
	  printf("Usage: %s [options] filename\n", progname);
	  puts("\
  -h,   --help                Print this help page and exit
  -V,   --version             Print version number and exit
  -o x,y --origo=x,y          image offset in cm, default 0,0
  -r x,y --rectangle=x,y      image size in cm, default 21.6,29.6
  -X dpi --x-resolution=dpi   dots per inch: only 300 works
  -Y dpi --y-resolution=dpi   dots per inch: only 300 works
  -b i  --brightness=i        0 auto, 1 dark, 7 bright, default 0
  -t i  --threshold=i         0 auto, 1 dark, 7 bright, default 4
  -c c  --image-composition=c 0 threshold, 1 dither, 2 grayscale, default 0
  -p t  --halftone=t          0 fine fatting, 1 scatter, 2 coarse fatting, 
                              3 mesh, default 0
  -x    --outline-extraction  
  -e    --edge-enhancement    
  -s    --separation          auto select between -c0 and -c1
  -m    --mirror              invert horizontal axis
  -i    --tiff                default is uncompressed PBM/PGM
  -v    --verbose             Give option twice for all details
Brightness and halftone are only used when dithering (-c1 or -s).");
	  
	  exit(0);

	case 'o':
	  sscanf(optarg, "%f,%f", &xcm, &ycm);
	  break;
	  
	case 'r':
	  sscanf(optarg, "%f,%f", &width, &height);
	  break;
	  
	case 'b':
	  brightness = atol(optarg);
	  if (brightness < 0 || brightness > 7)
	    err++;
	  break;
	case 't':
	  threshold = atol(optarg);
	  if (threshold < 0 || threshold > 7)
	    err++;
	  break;
	case 'c':
	  composition = atol(optarg);
	  if (composition < 0 || composition > 2)
	    err++;
	  break;
	case 'p':
	  halftone = atol(optarg);
	  if (halftone < 0 || halftone > 3)
	    err++;
	  break;
	  
	case 'x':
	  outline_extraction++;
	  break;
	case 'e':
	  edge_enhancement++;
	  break;
	case 's':
	  separation++;
	  break;
	case 'm':
	  mirroring++;
	  break;
	case 'i':
	  use_tiff++;
	  break;
	case 'X':
	  xresolution = atol(optarg);
	  break;
	case 'Y':
	  yresolution = atol(optarg);
	  break;
	default:
	  err++;
	  break;
	}
    }

  if (err > 0 || (optind != argc-1 && optind != argc))
    {
      fprintf(stderr, usage, progname);
      exit(1);
    }
  if (optind == argc && use_tiff)
    {
      fputs("Cannot write TIFF files to standard output\n", stderr);
      exit(1);
    }

  fd = open("/dev/rsst5", O_RDONLY);
  if (fd < 0)
    {
      perror("Opening /dev/rsst5");
      exit(1);
    }

  if (scan_image(fd, xcm, ycm, width, height, 
		 xresolution, yresolution,
		 lookup[brightness], lookup[threshold], composition,
		 halftone, outline_extraction, edge_enhancement,
		 separation, mirroring,
		 &image, &img_len, &img_bufsize,
		 &xpixels, &ypixels) != 0)
    {
      exit(1);
    }

  close(fd);

  if (use_tiff)
    {
      if (output_tiff(argv[optind], composition, separation,
		      image, img_len, xpixels, ypixels,
		      xresolution, yresolution, xcm, ycm) < 0)
	{
	  exit(1);
	}
    }
  else
    {
      if (output_pnm(optind == argc ? NULL : argv[optind], 
		     composition, image, img_len,
		     xpixels, ypixels) < 0)
	{
	  exit(1);
	}
    }

  if (image)
    free(image);

  exit(0);
}
