/* scanner - scan images from an M3191F2
   $Id: scanner.c,v 1.9 1994/07/10 08:43:23 ceder Exp $
   Copyright (C) 1994  Lysator Computer Association

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
   Originally written by Peter Eriksson.
   Heavily modified by Per Cederqvist.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/scsi/scsi.h>
#include <sys/scsi/impl/uscsi.h>

#include "scanner.h"

struct ready_cdb
{
    unsigned char opcode;
    unsigned char lun       : 3;
    unsigned char reserved1 : 5;
    unsigned char reserved2[3];
    unsigned char control_byte;
};

struct inquiry_cdb
{
    unsigned char opcode;
    unsigned char lun : 3;
    unsigned char reserved1 : 5;
    unsigned char evpd : 1;
    unsigned char reserved2 : 7;
    unsigned char vpdi;
    unsigned char allocation_length;
    unsigned char control_byte;
};

struct inquiry_reply
{
    unsigned char peripheral_qualifier : 3;
    unsigned char peripheral_device_type : 5;
    unsigned char rmb : 1;
    unsigned char device_type_qualifier : 7;
    unsigned char iso_version : 2;
    unsigned char ecma_version : 3;
    unsigned char ansi_version : 3;
    unsigned char aenc : 1;
    unsigned char reserved1 : 3;
    unsigned char response_data_format : 4;
    unsigned char additional_length;
    unsigned char reserved2[2];
    unsigned char reladr : 1;
    unsigned char wbus32 : 1;
    unsigned char wbus16 : 1;
    unsigned char sync : 1;
    unsigned char linked : 1;
    unsigned char cache : 1;
    unsigned char cmque : 1;
    unsigned char reserved3 : 1;
    unsigned char vendor_identification[8];
    unsigned char product_identification[16];
    unsigned char product_revision_level[4];
    unsigned char reserved4[60];
};


struct read_cdb
{
    unsigned char opcode;
    unsigned char lun : 3;
    unsigned char reserved1 : 4;
    unsigned char reladr : 1;
    unsigned char transfer_data_type;
    unsigned char reserved2;
    unsigned short transfer_identification __attribute__ ((packed)) ;
    unsigned char transfer_length[3] __attribute__ ((packed));
    unsigned char control_byte __attribute__ ((packed));
};


struct windowparams_cdb
{
    unsigned char opcode __attribute__ ((packed));
    unsigned char lun : 3 __attribute__ ((packed));
    unsigned char reserved1 : 5 __attribute__ ((packed));
    unsigned char reserved2[4] __attribute__ ((packed));
    unsigned char transfer_length[3] __attribute__ ((packed));
    unsigned char control_byte __attribute__ ((packed));
};


struct windowparams_data
{
    /* Header */
    unsigned char reserved0[6] __attribute__ ((packed));
    unsigned short window_descriptor_block_length __attribute__ ((packed));

    /* Descriptor Block */
    unsigned char window_identifier __attribute__ ((packed));
    unsigned char reserved1 : 7 __attribute__ ((packed));
    unsigned char auto_f : 1 __attribute__ ((packed));
    unsigned short x_resolution __attribute__ ((packed));
    unsigned short y_resolution __attribute__ ((packed));
    unsigned long upper_left_x __attribute__ ((packed));
    unsigned long upper_left_y __attribute__ ((packed));
    unsigned long width __attribute__ ((packed));
    unsigned long length __attribute__ ((packed));
    unsigned char brightness __attribute__ ((packed));
    unsigned char threshold __attribute__ ((packed));
    unsigned char contrast __attribute__ ((packed));
    unsigned char image_composition __attribute__ ((packed));
    unsigned char bit_per_pixel __attribute__ ((packed));
    unsigned char halftone_pattern[2] __attribute__ ((packed));
    unsigned char rif : 1 __attribute__ ((packed));
    unsigned char reserved2 : 4 __attribute__ ((packed));
    unsigned char padding_type : 3 __attribute__ ((packed));
    unsigned short bit_ordering __attribute__ ((packed));
    unsigned char compression_type __attribute__ ((packed));
    unsigned char compression_argument __attribute__ ((packed));
    unsigned char reserved3[6] __attribute__ ((packed));

    /* Cells (optional) */
    struct {
	char len;
	char function_code;
    } cells[4];
};

struct request_sense_cdb
{
    unsigned char opcode __attribute__ ((packed));
    unsigned char lun : 3 __attribute__ ((packed));
    unsigned char reserved1 : 5 __attribute__ ((packed));
    unsigned char reserved2 __attribute__ ((packed));
    unsigned char reserved3 __attribute__ ((packed));
    unsigned char allocation_length __attribute__ ((packed));
    unsigned char control_byte __attribute__ ((packed));
};

struct request_sense_reply
{
  unsigned char valid : 1 __attribute__ ((packed));
  unsigned char error_code : 7 __attribute__ ((packed));
  unsigned char segment_number __attribute__ ((packed));
  unsigned char fm : 1 __attribute__ ((packed));
  unsigned char eom : 1 __attribute__ ((packed));
  unsigned char ili : 1 __attribute__ ((packed));
  unsigned char reserved : 1 __attribute__ ((packed));
  unsigned char sense_key : 4 __attribute__ ((packed));
  unsigned char information_bytes[3] __attribute__ ((packed));
  unsigned char additional_sense_length __attribute__ ((packed));
};

struct reserve_release_unit_cdb
{
  unsigned char opcode __attribute__ ((packed));
  unsigned char lun : 3 __attribute__ ((packed));
  unsigned char tp : 1 __attribute__ ((packed));
  unsigned char tpid : 3 __attribute__ ((packed));
  unsigned char reserved1 : 1 __attribute__ ((packed));
  unsigned char reserved[3] __attribute__ ((packed));
  unsigned char control_byte __attribute__ ((packed));
};
  
/* Issue a REQUEST SENSE command (X`03', page 5-50).

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

static int
request_sense(int fd)
{
    struct uscsi_cmd sbuf;
    struct request_sense_cdb cbuf;
    struct request_sense_reply rbuf;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x03;
    cbuf.lun = 0;
    cbuf.allocation_length = sizeof(rbuf);

    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) &rbuf;
    sbuf.uscsi_buflen = sizeof(rbuf);

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;

    printf("--------------------\n");
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
      printf("Warning: ioctl failed: %s", strerror(errno));
    }

    if (rbuf.valid != 1)
      printf("Unexpected: Valid: %d\n", rbuf.valid);
    if (rbuf.error_code != 0x70)
      printf("Unexpected: Error code: %d", rbuf.error_code);
    if (rbuf.segment_number != 0)
      printf("Unexpected: Segment number: %d", rbuf.segment_number);
    if (rbuf.fm != 0)
      printf("Unexpected: FM: %d", rbuf.fm);
    printf("EOM: %d\n", rbuf.eom);
    printf("ILI: %d\n", rbuf.ili);
    switch(rbuf.sense_key)
      {
      case 0:
	printf("Sense key: NO SENSE\n");
	break;
      case 2:
	printf("Sense key: NOT READY\n");
	break;
      case 4:
	printf("Sense key: HARDWARE ERROR\n");
	break;
      case 5:
	printf("Sense key: ILLEGAL REQUEST\n");
	break;
      case 6:
	printf("Sense key: UNIT ATTENTION\n");
	break;
      case 0x0b:
	printf("Sense key: ABORTED COMMAND\n");
	break;
      default:
	printf("Unexpected: Sense key: %d\n", rbuf.sense_key);
	break;
      }
    if (rbuf.ili == 1 || 1)
      {
	long foo;
	
	foo = (0x10000 * rbuf.information_bytes[0]
	       + 0x100 * rbuf.information_bytes[1]
	       + 0x1 * rbuf.information_bytes[2]);
	/* Two's complement if negative. */
	if (foo & 0x800000)
	  {
	    foo ^= 0xffffff;
	    foo++;
	  }
	printf("Information bytes: %ld\n", foo);
      }
    if (rbuf.additional_sense_length != 0)
      printf("Unexpected: Additional sense length: %d\n", 
	     rbuf.additional_sense_length);

    printf("--------------------\n");

    return (sbuf.uscsi_status >> 1) & 31;
}


/* Issue a TEST UNIT READY command (X`00', page 5-57).

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

static int
scanner_ready(int fd)
{
    struct uscsi_cmd sbuf;
    struct ready_cdb cbuf;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x00;
    cbuf.lun = 0;


    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) 0;
    sbuf.uscsi_buflen = 0;

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;
    
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
	return -1;
    }

    return (sbuf.uscsi_status >> 1) & 31;
}

/* Issue a RESERVE UNIT command (X`16', page 5-54).

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

static int
scanner_reserve(int fd)
{
    struct uscsi_cmd sbuf;
    struct reserve_release_unit_cdb cbuf;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x16;
    cbuf.lun = 0;

    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) 0;
    sbuf.uscsi_buflen = 0;

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;
    
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
	return -1;
    }

    return (sbuf.uscsi_status >> 1) & 31;
}

/* Issue a RELEASE UNIT command (X`17', page 5-54).

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

static int
scanner_release(int fd)
{
    struct uscsi_cmd sbuf;
    struct reserve_release_unit_cdb cbuf;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x17;
    cbuf.lun = 0;

    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) 0;
    sbuf.uscsi_buflen = 0;

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;
    
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
	return -1;
    }

    return (sbuf.uscsi_status >> 1) & 31;
}


/* Issue an INQUIRY command (X`12', page 5-38).

   Prints some information to stdout.

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

int scanner_inquiry(int fd)
{
    struct uscsi_cmd sbuf;
    struct inquiry_cdb cbuf;
    struct inquiry_reply rbuf;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x12;
    cbuf.lun = 0;
    cbuf.evpd = 0;
    cbuf.vpdi = 0;
    cbuf.allocation_length = sizeof(rbuf);

    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) &rbuf;
    sbuf.uscsi_buflen = sizeof(rbuf);

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;
    
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
      return -1;
    }

    printf("Vendor..... %.8s\n", rbuf.vendor_identification);
    printf("Product.... %.16s\n", rbuf.product_identification);
    printf("Revision... %.4s\n", rbuf.product_revision_level);
    
    return (sbuf.uscsi_status >> 1) & 31;
}


/* Issue a DEFINE WINDOW PARAMETERS command (X`24', page 5-29).

   This sets up a single window for 300 dpi. x, y, width and lenght
   specify the rectangle that the window should cover (measured in
   1/1200 inch, origo is at the top left corner). 

   Brightness and threshold can be set to 0 for normal operation.

   Image composition can be
       0 (bi-level b&w,
       1 (dither/halftone b&w)
       2 (multi-level b&w (grayscale)).

   Halftone pattern can be
     0 (fine fatting),
     1 (scatter),
     2 (coarse fatting)
     3 (mesh).  
   Halftone pattern is only used if image compositoin is 1, or if
   separation is specified.

   Outline_extraction, edge_enhancement, separation and mirroring
   should be 0 or 1.  (Separation means that the scanner automatically
   chooses between image composition 0 and 1).

   Return values:
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */

static int
set_window_params(int fd,
		  long x, long y, /* 1/1200 inch */
		  long width, long length, /* 1/1200 inch */
		  int xresolution, int yresolution,
		  int brightness, /* 0 for normal */
		  int threshold, /* 0 for auto */
		  int composition,
		  int halftone,
		  int outline_extraction,
		  int edge_enhancement,
		  int separation,
		  int mirroring)
{
    struct uscsi_cmd sbuf;
    struct windowparams_cdb cbuf;
    struct windowparams_data dbuf;
    int cell = 0;

    memset(&cbuf, 0, sizeof(cbuf));
    
    cbuf.opcode = 0x24;
    cbuf.lun = 0;
    cbuf.transfer_length[0] = 0;
    cbuf.transfer_length[1] = 0;
    cbuf.transfer_length[2] = sizeof(dbuf)-sizeof(dbuf.cells);

    memset(&dbuf, 0, sizeof(dbuf));
    dbuf.window_descriptor_block_length = sizeof(dbuf)-8-sizeof(dbuf.cells);

    if (verbose > 1)
      fprintf(stderr, "Setting up for %dx%d dpi scan\n", 
	      xresolution, yresolution);

    dbuf.x_resolution = xresolution;
    dbuf.y_resolution = yresolution;
    dbuf.upper_left_x = x;
    dbuf.upper_left_y = y;
    dbuf.width = width;
    dbuf.length = length;
    dbuf.brightness = brightness;
    dbuf.threshold = threshold;
    dbuf.contrast = 0;
    dbuf.image_composition = composition;
    dbuf.bit_per_pixel = (composition == 2 ? 0x06 : 0x01);
    dbuf.halftone_pattern[0] = 0;
    dbuf.halftone_pattern[1] = halftone;
    dbuf.rif = 0;
    dbuf.padding_type = 0;
    dbuf.bit_ordering = 0;
    dbuf.compression_type = 0;
    dbuf.compression_argument = 0;

    if (outline_extraction)
    {
	dbuf.cells[cell].len = 1;
	dbuf.cells[cell].function_code = 0x01;
	cell++;
    }
    
    if (edge_enhancement)
    {
	dbuf.cells[cell].len = 1;
	dbuf.cells[cell].function_code = 0x02;
	cell++;
    }
    
    if (separation)
    {
	dbuf.cells[cell].len = 1;
	dbuf.cells[cell].function_code = 0x03;
	cell++;
    }
    
    if (mirroring)
    {
	dbuf.cells[cell].len = 1;
	dbuf.cells[cell].function_code = 0x04;
	cell++;
    }
    
    cbuf.transfer_length[2] += 2*cell;
    dbuf.window_descriptor_block_length += 2*cell;

    sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_WRITE+
	USCSI_NODISCON+USCSI_NOTAG;

    sbuf.uscsi_status = 0;

    sbuf.uscsi_timeout = 1000;

    sbuf.uscsi_cdb = (caddr_t) &cbuf;
    sbuf.uscsi_cdblen = sizeof(cbuf);
    
    sbuf.uscsi_bufaddr = (caddr_t) &dbuf;
    sbuf.uscsi_buflen = cbuf.transfer_length[2];

    sbuf.uscsi_resid = 0;
    sbuf.uscsi_rqlen = 0;
    sbuf.uscsi_rqstatus = 0;
    sbuf.uscsi_rqresid = 0;
    sbuf.uscsi_rqbuf = 0;
    
    if (ioctl(fd, USCSICMD, &sbuf) < 0)
    {
	perror("setting window params: ioctl(USCSICMD)");
	exit(1);
    }

    return (sbuf.uscsi_status >> 1) & 31;
}



/* Issue a READ command (X`28', page 5-42).

   set_window_params must already have been used.  This will cause the
   scanner to scan, and read back the result.  The result will be
   written to filename as a PBM bitmap (if composition is 0 or 1) or
   PGM grayscale format (if composition is 2).

   You should specify the same arguments to this function as to the
   previous set_window_params.

   Return values:
     -2  Failed to allocate memory for the result.
     -1  The ioctl failed.  errno holds an error number.
      0  GOOD
      2  CHECK CONDITION (send a REQUEST SENSE to find out what is wrong)
      8  BUSY  (wait a while and try again)
   0x18  RESERVATION CONFLICT

   Other return values indicates that the scanner returns undocumented
   results. */
static int
scanner_read(int fd,
	     long x, long y,
	     long width, long length,
	     int xresolution, int yresolution,
	     int composition,
	     char **image,	/* in/out */
	     long *img_len,	/* out */
	     long *img_bufsize,	/* in/out */
	     long *x_pixels,	/* out */
	     long *y_pixels)	/* out */
{
    struct uscsi_cmd sbuf;
    struct read_cdb cbuf;
    long lines;			/* Number of scan lines in the image */
    long block;			/* Number of bytes per scan line */
    char *chunkstart;

    if (verbose > 1)
      fprintf(stderr, "Reading %dx%d dpi image\n", xresolution, yresolution);

    if (composition == 2)
      {
	/* Multi-level B&W */
	block = xresolution * width / 1200;
	if (x_pixels != NULL)
	  *x_pixels = block;
      }
    else
      {
	/* Bi-level B&W, Dither/halftone B&W */
	block = (xresolution * width / 1200 + 7) / 8;
	if (x_pixels != NULL)
	  *x_pixels = block*8;
      }

    lines = yresolution * length / 1200;
    if (y_pixels != NULL)
      *y_pixels = lines;
    *img_len = block * lines;

    if (*image == NULL || *img_bufsize == 0)
    {
	*image = malloc(*img_len);
	*img_bufsize = *img_len;
    }
    else if (*img_len > *img_bufsize)
    {
	*image = realloc(*image, *img_len);
	*img_bufsize = *img_len;
    }
    if (*image == NULL)
	return -2;

    chunkstart = *image;
    while (lines > 0)
      {
	long chunk = 100000 / block;
	if (chunk > lines)
	  chunk = lines;

	if (verbose > 1)
	{
	    fprintf(stderr, "Receiving %ld-sized chunk, lines=%ld\n",
		    chunk, lines);
	}
	
	memset(&cbuf, 0, sizeof(cbuf));
    
	cbuf.opcode = 0x28;
	cbuf.lun = 0;
	cbuf.reladr = 0;
	cbuf.transfer_data_type = 0;
	cbuf.transfer_identification = 0x00;
	cbuf.transfer_length[0] = chunk/0x10000;
	cbuf.transfer_length[1] = (chunk/0x100) & 0xff;
	cbuf.transfer_length[2] = chunk & 0xff;
    
	sbuf.uscsi_flags = USCSI_DIAGNOSE+USCSI_ISOLATE+USCSI_READ+
	  /* USCSI_NODISCON+ */ USCSI_NOTAG;

	sbuf.uscsi_status = 0;

	sbuf.uscsi_timeout = 1000;

	sbuf.uscsi_cdb = (caddr_t) &cbuf;
	sbuf.uscsi_cdblen = sizeof(cbuf);
    
	sbuf.uscsi_bufaddr = (caddr_t) chunkstart;
	sbuf.uscsi_buflen = block * chunk;

	sbuf.uscsi_resid = 0;
	sbuf.uscsi_rqlen = 0;
	sbuf.uscsi_rqstatus = 0;
	sbuf.uscsi_rqresid = 0;
	sbuf.uscsi_rqbuf = 0;
    
	if (ioctl(fd, USCSICMD, &sbuf) < 0)
	  {
	    perror("reading image: ioctl(USCSICMD)");
	    printf("Request sense result: %d\n", request_sense(fd));
	    if ((sbuf.uscsi_status >> 1) & 31 != 0)
	      return (sbuf.uscsi_status >> 1) & 31 != 0;
	    else
	      return -1;
	  }

	lines -= chunk;
	chunkstart += block * chunk;
	
	if ((sbuf.uscsi_status >> 1) & 31 != 0)
	  return (sbuf.uscsi_status >> 1) & 31 != 0;
      }
    return 0;
}

int
scan_image(int fd,
	   float xcm,		/* cm */
	   float ycm,		/* cm */
	   float widthcm,	/* cm */
	   float lengthcm,	/* cm */
	   int xresolution,	/* dpi */
	   int yresolution,	/* dpi */
	   int brightness,
	   int threshold,
	   int composition,
	   int halftone,
	   int outline_extraction,
	   int edge_enhancement,
	   int separation,
	   int mirroring,
	   char **image,
	   long *img_len,
	   long *img_bufsize,
	   long *xpixels,
	   long *ypixels)
{
    int code;
    long x, y;
    long width, length;

    if (verbose)
	fprintf(stderr, "Scanning %f cm x %f cm %s image\n",
		widthcm, lengthcm, composition == 2 ? "grayscale" : "b/w");

    if (sizeof(struct windowparams_cdb) != 10)
    {
	fprintf(stderr, "Illegal sizeof(windowparams_cdb): %d\n",
	       sizeof(struct windowparams_cdb) );
	return -1;
    }

    x = xcm / 2.54 * 1200;
    y = ycm / 2.54 * 1200;
    if (x < 0 || x > 10204)
      {
	fprintf(stderr, "Bad x value");
	return -1;
      }
    if (y < 0 || y > 14028)
      {
	fprintf(stderr, "Bad y value");
	return -1;
      }

    width = widthcm / 2.54 * 1200;
    length = lengthcm / 2.54 * 1200;
    if (width < 0 || length < 0)
      {
	fprintf(stderr, "Negative sizes not supported");
	return -1;
      }
    if (x + width > 10204 || y + length > 14028)
      {
	fprintf(stderr, "Attempt to scan outside paper\n");
	return -1;
      }
    
    while ((code = scanner_ready(fd)) == 8)
    {
	fprintf(stderr, "Scanner is busy...\n");
	sleep(1);
    }
    if (code != 0)
    {
	fprintf(stderr, "Bad code from scanner: %d\n", code);
	return -1;
    }

    if (verbose > 1)
    {
	sleep(1);
	fprintf(stderr, "Reserving scanner\n");
    }

    if ((code = scanner_reserve(fd)) != 0)
    {
	fprintf(stderr, "Bad code reserving scanner: %d\n", code);
	return -1;
    }

    if (verbose > 1)
    {
	sleep(1);
	fprintf(stderr, "Setting window params\n");
    }

    while ((code = set_window_params(fd, x, y, width, length,
				     xresolution, yresolution,
				     brightness, threshold,
				     composition, halftone,
				     outline_extraction,
				     edge_enhancement,
				     separation, mirroring)) == 8)
    {
	fprintf(stderr, "set_window_params: Scanner is busy...\n");
	sleep(1);
    }
    if (code != 0)
    {
	fprintf(stderr, "set_window_params:  Bad code from scanner: %d", code);
	return -1;
    }

    while ((code = scanner_read(fd, x, y, width, length, 
				xresolution, yresolution, composition,
				image, img_len, img_bufsize,
				xpixels, ypixels)) == 8)
    {
	fprintf(stderr, "scanner_read: Scanner is busy...\n");
	sleep(1);
    }

    if (code != 0)
    {
	fprintf(stderr, "scanner_read:  Bad code from scanner: %d", code);
	return -1;
    }

#if 0 /* Doesn't work, and crashes scanner */
    if (verbose > 1)
    {
	sleep(1);
	fprintf(stderr, "Releasing scanner\n");
    }

    if ((code = scanner_release(fd)) != 0)
    {
	fprintf(stderr, "Bad code releasing scanner: %d\n", code);
	return -1;
    }
#endif

    return 0;
}
